const { wrapTimeout } = require('@consento/promise/wrapTimeout')
const { bubbleAbort } = require('@consento/promise/bubbleAbort')
const DNS_PREFIX = '_dnslink.'

function dnslink (domain, options) {
  return wrapTimeout(signal => {
    const validated = validateDomain(domain)
    if (validated.error) {
      return { links: {}, path: [], log: [validated.error] }
    }
    return dnslinkN(validated.redirect, { ...options, signal }, [])
      .then(result => {
        result.path = []
        for (const entry of result.log.filter(entry => entry.code === 'REDIRECT' || entry.code === 'RESOLVE')) {
          if (entry.pathname || entry.search) {
            result.path.unshift({
              pathname: entry.pathname,
              search: entry.search
            })
          }
        }
        return result
      })
  }, options)
}

dnslink.LogCode = {
  redirect: 'REDIRECT',
  resolve: 'RESOLVE',
  conflictEntry: 'CONFLICT_ENTRY',
  invalidEntry: 'INVALID_ENTRY',
  endlessRedirect: 'ENDLESS_REDIRECT',
  invalidRedirect: 'INVALID_REDIRECT',
  tooManyRedirects: 'TOO_MANY_REDIRECTS',
  unusedEntry: 'UNUSED_ENTRY',
  recursivePrefix: 'RECURSIVE_DNSLINK_PREFIX'
}

dnslink.InvalidityReason = {
  wrongStart: 'WRONG_START',
  keyMissing: 'KEY_MISSING',
  noValue: 'NO_VALUE'
}

module.exports = dnslink

function shouldFallbackToDomain (result) {
  // Any log entry given already prevents the fallback from _dnslink.<domain> to <domain>
  if (result.log.length > 0) {
    return false
  }
  /* eslint-disable-next-line no-unreachable-loop */
  for (const _key in result.links) {
    // Any result will also prevent the fallback
    return false
  }
  return true
}

async function dnslinkN (source, options, chain) {
  const { domain } = source
  let result = await resolveDnslink(domain, options)
  bubbleAbort(options.signal)
  let redirect
  if (result.links.dns) {
    const validated = validateDomain(result.links.dns.value)
    if (validated.error) {
      result.log.push(validated.error)
    } else {
      redirect = validated.redirect
    }
  } else if (domain.startsWith(DNS_PREFIX) && shouldFallbackToDomain(result)) {
    redirect = {
      domain: domain.substr(DNS_PREFIX.length)
    }
  }
  const resolve = { code: 'RESOLVE', ...source }
  if (redirect) {
    const log = extractRedirectLog(result)
    if (chain.includes(redirect.domain)) {
      log.push(resolve)
      return { links: {}, log: [...log, { code: 'ENDLESS_REDIRECT', ...redirect }] }
    }
    chain.push(domain)
    if (chain.length === 32) {
      log.push(resolve)
      return { links: {}, log: [...log, { code: 'TOO_MANY_REDIRECTS', ...redirect }] }
    }
    log.push({ code: 'REDIRECT', ...source })
    result = await dnslinkN(redirect, options, chain)
    result.log = log.concat(result.log)
    return result
  }
  const { links: linksRaw, log } = result
  const links = {}
  for (const key in linksRaw) {
    links[key] = linksRaw[key].value
  }
  log.push(resolve)
  return { links, log }
}

function extractRedirectLog (result) {
  const { log, links } = result
  for (const key in links) {
    if (key === 'dns') continue
    log.push({ code: 'UNUSED_ENTRY', entry: links[key].entry })
  }
  return log
}

const PREFIX = 'dnslink='

async function resolveDnslink (domain, options) {
  const txtEntries = (await resolveTxt(domain, options))
    .reduce((combined, array) => combined.concat(array), [])

  const log = []
  const links = {}
  for (const entry of txtEntries) {
    if (!entry.startsWith(PREFIX)) {
      continue
    }
    const validated = validate(entry)
    if (validated.error !== undefined) {
      log.push({ code: 'INVALID_ENTRY', entry, reason: validated.error })
      continue
    }
    const { key, value } = validated
    const prev = links[key]
    if (!prev || prev.value > value) {
      if (prev) {
        log.push({ code: 'CONFLICT_ENTRY', entry: prev.entry })
      }
      links[key] = {
        value,
        entry
      }
    } else {
      log.push({ code: 'CONFLICT_ENTRY', entry })
    }
  }

  return { links, log }
}

function validateDomain (input) {
  let { domain, pathname, search } = relevantURLParts(input)
  if (domain.startsWith(DNS_PREFIX)) {
    domain = domain.substr(DNS_PREFIX.length)
    if (domain.startsWith(DNS_PREFIX)) {
      return { error: { code: 'RECURSIVE_DNSLINK_PREFIX', domain: `${DNS_PREFIX}${domain}`, pathname, search } }
    }
  }
  if (domain.includes(' ')) {
    return { error: { code: 'INVALID_REDIRECT', domain, pathname, search } }
  }
  return { redirect: { domain: `${DNS_PREFIX}${domain}`, pathname, search } }
}

function relevantURLParts (input) {
  const url = new URL(input, 'xo://')
  let domain
  let pathname
  if (url.hostname) {
    domain = url.hostname
    if (url.pathname) {
      pathname = url.pathname
    }
  } else if (url.pathname) {
    const parts = /^\/([^/]*)(\/.*)?/.exec(url.pathname)
    domain = parts[1]
    pathname = parts[2]
  }
  let search
  for (const key of url.searchParams.keys()) {
    if (search === undefined) {
      search = {}
    }
    search[key] = url.searchParams.getAll(key)
  }
  return { search, domain, pathname }
}

function validate (entry) {
  const trimmed = entry.substr(PREFIX.length).trim()
  if (!trimmed.startsWith('/')) {
    return { error: 'WRONG_START' }
  }
  const parts = trimmed.split('/')
  parts.shift()
  let key
  if (parts.length !== 0) {
    key = parts.shift().trim()
  }
  if (!key) {
    return { error: 'KEY_MISSING' }
  }
  let value
  if (parts.length !== 0) {
    value = parts.join('/').trim()
  }
  if (!value) {
    return { error: 'NO_VALUE' }
  }
  return { key, value }
}

let dohQuery
let dnsQuery

async function resolveTxt (domain, options) {
  const { doh, dns } = options
  if (!dns || (doh && Math.random() > 0.5)) {
    if (dohQuery === undefined) {
      dohQuery = require('doh-query').query
    }
    const { answers } = await dohQuery({ questions: [{ type: 'TXT', name: domain }] }, {
      signal: options.signal,
      endpoints: doh
    })
    return answers
      .map(entry => Object.values(entry.data).map(entry => entry.toString()))
  }
  if (dnsQuery === undefined) {
    dnsQuery = require('./query-txt.node.js')
  }
  return dnsQuery(domain, options)
}
